@@ -14,6 +14,8 @@ from numpyencoder import NumpyEncoder
 
 __all__ = ['TimeSeriesEncoder', 'JSONEncoder', 'CSVEncoder']
 
+MAX_FLOATING_PRECISION = 6
+
 class EncoderHelpers:
     @staticmethod
     def precision_and_scale_np(x, max_magnitude):
@@ -49,13 +51,18 @@ class EncoderHelpers:
 
     @staticmethod
     def calculate_bit_depth(values, encoding_size):
-        longest_input_length = np.max(np.vectorize(len)(values.astype(np.str_)))
-        maximum_precision = EncoderHelpers.precision_and_scale_np(values, longest_input_length)
-
         max_value = np.max(values)
         min_value = np.min(values)
-
         max_value = max(abs(max_value), abs(min_value))
+
+        vals = np.copy(values)
+        maximum_precision = 0
+        while np.all(np.rint(vals) == vals) == False:
+            maximum_precision += 1
+            if maximum_precision >= MAX_FLOATING_PRECISION:
+                break
+            vals *= 10
+
         if maximum_precision != 0:
             numeric_type = "float"
             max_value *= 10 ** maximum_precision
@@ -79,8 +86,7 @@ class EncoderHelpers:
         num_vals = len(values)
         encoding_depth, max_prec, num_type, signed = EncoderHelpers.calculate_bit_depth(np.arange(1, num_vals+1), encoding_size=encoding_size)
         encoder = NumericEncoder(numeric_type=num_type, signed=signed, float_precision=max_prec, encoding_depth=encoding_depth, encoding_size=encoding_size)
-        encoded = encoder.encode(np.arange(0, num_vals))
-        encoded_states = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
+        encoded_states = encoder.encode(np.arange(0, num_vals), joined=False)
         lookup = {}
         for i, s in enumerate(values):
             lookup[s] = encoded_states[i]
@@ -308,12 +314,11 @@ class JSONEncoder(TimeSeriesEncoder):
             
     @staticmethod
     def decode_json(json_data, inplace=False, gzip=False):
+        if inplace == False:
+            json_data = copy.copy(json_data)
         if gzip:
             json_data = EncoderHelpers.gunzip_bytes_obj(json_data)
             json_data = json.loads(json_data)
-
-        if inplace == False and gzip == False:
-            json_data = copy.copy(json_data)
         decoded = JSONEncoder._decode_json(json_data)
         return decoded
 
@@ -426,56 +431,54 @@ class CSVEncoder(TimeSeriesEncoder):
             try:
                 float(df[col][0])
                 deci_count = len(df[col][0].split('.')[1])
-                self.value_columns[col]["format"] = f'.{deci_count}f'
+                self.value_columns[col]["format"] = f'.{min(deci_count, MAX_FLOATING_PRECISION)}f'
             except:
                 continue
             
 
     def encode_time(self, df, time_column):
-
-        times = [x.timestamp() for x in np.vectorize(ciso8601.parse_datetime)(df[time_column].values)]
+        times = pd.to_datetime(df[time_column]).astype(int) / 10**9
         self.times = times
-        self._set_time_params(col_name=time_column, start=min(times))
-        gaps = np.insert(np.diff(times), 0, 0).astype(np.int64)
+        self._set_time_params(col_name=time_column, start=times.min())
+        gaps = np.insert(np.diff(times.to_numpy()), 0, 0).astype(np.int64)
 
         # Calculate encoder params
         encoding_depth, max_prec, num_type, signed = EncoderHelpers.calculate_bit_depth(gaps, encoding_size=self.encoding_size)
+
         # Do direct encoding
         encoder = NumericEncoder(numeric_type=num_type, signed=signed, float_precision=max_prec, encoding_depth=encoding_depth, encoding_size=self.encoding_size)
-        encoded = encoder.encode(gaps)
         self._set_time_params(encoder=encoder)
-        words = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
+        words = encoder.encode(gaps, joined=False)
 
         # Decided if we encode the values directly, or use a lookup table
         states = set(words)
         num_states = len(states)
         str_len_states = len(str(states))
         lookup_bit_depth = EncoderHelpers._calculate_bit_depth(num_states, self.encoding_size)
-
-        if lookup_bit_depth * len(words) + str_len_states < len(encoded):
+        if lookup_bit_depth * len(words) + str_len_states < len(words) * encoding_depth:
             # Do lookup table
-            lookup, encoding_depth = EncoderHelpers.create_lookup_table(words)
-            encoded = ''
-            for i in words:
-                encoded += lookup[i]
+            lookup, encoding_depth = EncoderHelpers.create_lookup_table(states)
+            encoded = list(map(lookup.get, words))
             self._set_time_params(lookup=lookup)
-            words = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
-
+            words = encoded
         return np.asarray(words).reshape(-1, 1)
     
     def encode_keys(self, df, key_columns):
         #Build unique aggregate key
         df = df[key_columns]
-        aggregate_keys = df[key_columns].astype(str).agg('|'.join, axis=1)
+        aggregate_keys = None
+        for c in df.columns:
+            if aggregate_keys is None:
+                aggregate_keys = df[c]
+            else:
+                aggregate_keys += "|" + df[c]
 
         lookup, encoding_depth = EncoderHelpers.create_lookup_table(aggregate_keys, self.encoding_size)
         self._set_key_params(col_names=key_columns, lookup=lookup)
-        encoded = ''
-        for i in aggregate_keys:
-            encoded += lookup[i]
-
-        words = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
-        return np.asarray(words).reshape(-1, 1)
+        encoded = [0] * len(aggregate_keys)
+        for i, v in enumerate(aggregate_keys):
+            encoded[i] = lookup[v]
+        return np.asarray(encoded).reshape(-1, 1)
     
     def encode_value(self, df, value_column):
         vals = df[value_column].values
@@ -497,9 +500,8 @@ class CSVEncoder(TimeSeriesEncoder):
             encoding_depth, max_prec, num_type, signed = EncoderHelpers.calculate_bit_depth(vals, encoding_size=self.encoding_size)
             # Do direct encoding
             encoder = NumericEncoder(numeric_type=num_type, signed=signed, float_precision=max_prec, encoding_depth=encoding_depth, encoding_size=self.encoding_size)
-            encoded = encoder.encode(vals)
+            words = encoder.encode(vals, joined=False)
             self._set_encoded_column(column_name=value_column, encoder=encoder)
-            words = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
         else:
             words = vals
             encoded = ''.join(vals)
@@ -509,20 +511,18 @@ class CSVEncoder(TimeSeriesEncoder):
         num_states = len(states)
         str_len_states = len(str(states))
         lookup_bit_depth = EncoderHelpers._calculate_bit_depth(num_states, self.encoding_size)
-        if lookup_bit_depth * len(words) + str_len_states < len(encoded):
+        if lookup_bit_depth * len(words) + str_len_states < len(words) * encoding_depth:
             # Do lookup table
-            lookup, encoding_depth = EncoderHelpers.create_lookup_table(words)
-            encoded = ''
-            for i in words:
-                encoded += lookup[i]
+            lookup, encoding_depth = EncoderHelpers.create_lookup_table(states)
+            encoded = list(map(lookup.get, words))
             self._set_lookup_column(column_name=value_column, lookup=lookup)
-            words = [encoded[i:i+encoding_depth] for i in range(0, len(encoded), encoding_depth)]
-        
+            words = encoded
         return np.asarray(words).reshape(-1, 1)
 
     @staticmethod
     def encode_csv(csv, time_column, key_columns, sort_values = True, encoding_size = 64, gzip=False, functional_compression=False):
         df = pd.read_csv(StringIO(csv))
+        
         if sort_values:
             df = df.sort_values(time_column, ascending=True)
 
@@ -531,6 +531,7 @@ class CSVEncoder(TimeSeriesEncoder):
 
         ndf = pd.DataFrame(encoder.encode_time(df, time_column), columns=[time_column])
         ndf["keys"] = encoder.encode_keys(df, key_columns)
+        
         tscols = set(df.columns) - set([time_column] + key_columns)
         for col in tscols:
             encoded = encoder.encode_value(df, value_column=col)
@@ -539,10 +540,14 @@ class CSVEncoder(TimeSeriesEncoder):
                 ndf[col] = encoded
         
         encoder._set_value_column_fmt(csv)
-
         packet = encoder.__dict__
         del packet["times"]
-        data = ndf.astype(str).agg(''.join, axis=1)
+        data = None
+        for c in ndf.columns:
+            if data is None:
+                data = ndf[c]
+            else:
+                data += ndf[c]
         packet["data"] = ''.join(data)
         encoded = json.dumps(packet)
         if gzip:
@@ -586,16 +591,15 @@ class CSVEncoder(TimeSeriesEncoder):
         if 'lookup' in time_vals:
             lookup = time_vals["lookup"]
             lookup =  {v: k for k, v in lookup.items()}
-            tokens = np.vectorize(lookup.get)(tokens)
+            tokens = list(map(lookup.get, tokens))
 
         if 'encoder' in time_vals:
             encoder = NumericEncoder.deserialize(time_vals["encoder"])
             tokens = encoder.decode(''.join(tokens))
-        
         cumulative_time = np.cumsum(np.asarray(tokens))
         time = cumulative_time + start
         self.time = time
-        fmttimes = np.vectorize(datetime.datetime.utcfromtimestamp)(time)
+        fmttimes = list(map(datetime.datetime.utcfromtimestamp, time))
         fmttimes = ['%02d-%02d-%02dT%02d:%02d:%02dZ' % (x.year, x.month, x.day, x.hour, x.minute, x.second) for x in fmttimes]
         return fmttimes
 
@@ -603,10 +607,8 @@ class CSVEncoder(TimeSeriesEncoder):
     def decode_key(self, json_data, tokens):
         columns = json_data["keys"]["columns"]
         lookup = json_data["keys"]["lookup"]
-        lookup =  {v: k for k, v in lookup.items()}
-        new_tokens = np.vectorize(lookup.get)(tokens)
-        new_tokens = np.char.split(new_tokens, sep="|")
-        new_tokens = np.array([x for x in new_tokens])
+        lookup =  {v: k.split('|') for k, v in lookup.items()}
+        new_tokens = list(map(lookup.get, tokens))
         key_data = pd.DataFrame(new_tokens, columns=columns)
         return key_data
 
@@ -614,6 +616,14 @@ class CSVEncoder(TimeSeriesEncoder):
         value_columns = json_data["value_columns"]
         df = pd.DataFrame(np.ones((len(tokens), len(json_data["value_columns"]))))
         df.columns = [col for col in value_columns]
+        def parse_col_tokens(col_bytes, tokens):
+            col_vals = [0] * len(tokens)
+            new_tokens = [0] * len(tokens)
+            for i, t in enumerate(tokens):
+                col_vals[i] = t[:col_bytes]
+                new_tokens[i] = t[col_bytes:]
+            return col_vals, new_tokens
+
         for col in value_columns:
             if "function" in value_columns[col]:
                 coefs = value_columns[col]["function"]
@@ -630,34 +640,30 @@ class CSVEncoder(TimeSeriesEncoder):
                     lookup = col_json["lookup"]
                     lookup =  {v: k for k, v in lookup.items()}
                     token_size = len(list(lookup.keys())[0])
-                    tokens = [(x[:token_size], x[token_size:]) for x in tokens]
-                    col_tokens = np.vectorize(lookup.get)([x[0] for x in tokens])
+                    col_vals, tokens = parse_col_tokens(token_size, tokens)
+                    col_tokens = list(map(lookup.get, col_vals))
                     if 'encoder' in col_json:
                         encoder = NumericEncoder.deserialize(col_json["encoder"])
                         col_tokens = encoder.decode(''.join(col_tokens))
                     df[col] = col_tokens
-                    tokens = [x[1] for x in tokens]
                 elif 'encoder' in col_json:
                     encoder = NumericEncoder.deserialize(col_json["encoder"])
                     token_size = encoder.encoding_depth
-                    tokens = [(x[:token_size], x[token_size:]) for x in tokens]
-                    col_tokens = encoder.decode(''.join([x[0] for x in tokens]))
+                    col_vals, tokens = parse_col_tokens(token_size, tokens)
+                    col_tokens = encoder.decode(''.join(col_vals))
                     df[col] = col_tokens
-                    tokens = [x[1] for x in tokens]
 
             if "format" in value_columns[col]:
                 fmt = value_columns[col]["format"]
-                df[col] = df[col].apply(lambda x: f"{x:{fmt}}")
-            
-
+                df[col] = df[col].map(f"{{:{fmt}}}".format)
         return df
 
     def tokenize(self, data, time_size, key_size, value_size):
         num_tokens = len(data) / (time_size + key_size + value_size)
         num_tokens = int(num_tokens)
-        times = np.zeros(num_tokens).astype(np.str_)
-        keys = np.zeros(num_tokens).astype(np.str_)
-        values = np.zeros(num_tokens).astype(np.str_)
+        times = [0] * num_tokens
+        keys = [0] * num_tokens
+        values = [0] * num_tokens
         for token, i in enumerate(range(0, len(data), (time_size + key_size + value_size))):
             times[token] = data[i:i+time_size]
             keys[token] = data[i+time_size:i+time_size+key_size]
@@ -678,7 +684,14 @@ class CSVEncoder(TimeSeriesEncoder):
         ndf = ndf.join(decoder.decode_key(json_data, keys))
         ndf = ndf.join(decoder.decode_values(json_data, values))
         ndf = ndf[json_data["columns"]]
-        return ndf.to_csv(index=False)
+
+        rslt = None
+        for c in ndf.columns:
+            if rslt is None:
+                rslt = ndf[c]
+            else:
+                rslt += ',' + ndf[c]
+        return ','.join(ndf.columns) + '\n' + '\n'.join(rslt.values)
 
     def __init__(self, encoding_size=64, functional_compression=False):
         self.encoding_size = encoding_size
